@@ -8,22 +8,30 @@ import (
 	"strings"
 	"time"
 
+	"github.com/jandedobbeleer/oh-my-posh/src/log"
 	"github.com/jandedobbeleer/oh-my-posh/src/properties"
 	"github.com/jandedobbeleer/oh-my-posh/src/regex"
 	"github.com/jandedobbeleer/oh-my-posh/src/runtime"
+	"github.com/jandedobbeleer/oh-my-posh/src/runtime/path"
 
 	"gopkg.in/ini.v1"
 )
 
 type Commit struct {
-	// git log -1 --pretty="format:%an%n%ae%n%cn%n%ce%n%at%n%s"
+	Timestamp time.Time
 	Author    *User
 	Committer *User
+	Refs      *Refs
 	Subject   string
-	Timestamp time.Time
 	Sha       string
 }
 
+type Refs struct {
+	Heads   []string
+	Tags    []string
+	Remotes []string
+}
+
 type User struct {
 	Name  string
 	Email string
@@ -111,6 +119,8 @@ const (
 	UntrackedModes properties.Property = "untracked_modes"
 	// IgnoreSubmodules list the optional ignore-submodules mode per repo
 	IgnoreSubmodules properties.Property = "ignore_submodules"
+	// MappedBranches allows overriding certain branches with an icon/text
+	MappedBranches properties.Property = "mapped_branches"
 
 	DETACHED     = "(detached)"
 	BRANCHPREFIX = "ref: refs/heads/"
@@ -119,38 +129,41 @@ const (
 	trueStr = "true"
 )
 
-type Git struct {
-	scm
+type Rebase struct {
+	HEAD    string
+	Onto    string
+	Current int
+	Total   int
+}
 
+type Git struct {
+	User           *User
 	Working        *GitStatus
 	Staging        *GitStatus
-	Ahead          int
-	Behind         int
-	HEAD           string
+	commit         *Commit
+	Rebase         *Rebase
+	RawUpstreamURL string
 	Ref            string
 	Hash           string
 	ShortHash      string
 	BranchStatus   string
 	Upstream       string
+	HEAD           string
 	UpstreamIcon   string
 	UpstreamURL    string
-	RawUpstreamURL string
-	UpstreamGone   bool
-	IsWorkTree     bool
-	IsBare         bool
-	User           *User
-	Detached       bool
-	Merge          bool
-	Rebase         bool
-	CherryPick     bool
-	Revert         bool
-
-	// needed for posh-git support
-	poshgit       bool
-	stashCount    int
+	scm
 	worktreeCount int
-
-	commit *Commit
+	stashCount    int
+	Behind        int
+	Ahead         int
+	IsWorkTree    bool
+	Merge         bool
+	CherryPick    bool
+	Revert        bool
+	poshgit       bool
+	Detached      bool
+	IsBare        bool
+	UpstreamGone  bool
 }
 
 func (g *Git) Template() string {
@@ -158,6 +171,7 @@ func (g *Git) Template() string {
 }
 
 func (g *Git) Enabled() bool {
+	// g.command = GITCOMMAND
 	g.User = &User{}
 
 	if !g.shouldDisplay() {
@@ -179,7 +193,8 @@ func (g *Git) Enabled() bool {
 		return true
 	}
 
-	if g.hasPoshGitStatus() {
+	source := g.props.GetString(Source, Cli)
+	if source == Pwsh && g.hasPoshGitStatus() {
 		return true
 	}
 
@@ -187,6 +202,7 @@ func (g *Git) Enabled() bool {
 	if g.shouldIgnoreStatus() {
 		displayStatus = false
 	}
+
 	if displayStatus {
 		g.setGitStatus()
 		g.setGitHEADContext()
@@ -194,21 +210,37 @@ func (g *Git) Enabled() bool {
 	} else {
 		g.setPrettyHEADName()
 	}
+
 	if g.props.GetBool(FetchUpstreamIcon, false) {
 		g.UpstreamIcon = g.getUpstreamIcon()
 	}
+
 	return true
 }
 
+func (g *Git) CacheKey() (string, bool) {
+	dir, err := g.env.HasParentFilePath(".git", true)
+	if err != nil {
+		return "", false
+	}
+
+	ref := g.FileContents(dir.Path, "HEAD")
+	ref = strings.Replace(ref, "ref: refs/heads/", "", 1)
+	return fmt.Sprintf("%s@%s", dir.Path, ref), true
+}
+
 func (g *Git) Commit() *Commit {
 	if g.commit != nil {
 		return g.commit
 	}
+
 	g.commit = &Commit{
 		Author:    &User{},
 		Committer: &User{},
+		Refs:      &Refs{},
 	}
-	commitBody := g.getGitCommandOutput("log", "-1", "--pretty=format:an:%an%nae:%ae%ncn:%cn%nce:%ce%nat:%at%nsu:%s%nha:%H")
+
+	commitBody := g.getGitCommandOutput("log", "-1", "--pretty=format:an:%an%nae:%ae%ncn:%cn%nce:%ce%nat:%at%nsu:%s%nha:%H%nrf:%D", "--decorate=full")
 	splitted := strings.Split(strings.TrimSpace(commitBody), "\n")
 	for _, line := range splitted {
 		line = strings.TrimSpace(line)
@@ -234,6 +266,25 @@ func (g *Git) Commit() *Commit {
 			g.commit.Subject = line
 		case "ha:":
 			g.commit.Sha = line
+		case "rf:":
+			refs := strings.Split(line, ", ")
+			for _, ref := range refs {
+				ref = strings.TrimSpace(ref)
+				switch {
+				case strings.HasSuffix(ref, "HEAD"):
+					continue
+				case strings.HasPrefix(ref, "tag: refs/tags/"):
+					g.commit.Refs.Tags = append(g.commit.Refs.Tags, strings.TrimPrefix(ref, "tag: refs/tags/"))
+				case strings.HasPrefix(ref, "refs/remotes/"):
+					g.commit.Refs.Remotes = append(g.commit.Refs.Remotes, strings.TrimPrefix(ref, "refs/remotes/"))
+				case strings.HasPrefix(ref, "HEAD -> refs/heads/"):
+					g.commit.Refs.Heads = append(g.commit.Refs.Heads, strings.TrimPrefix(ref, "HEAD -> refs/heads/"))
+				case strings.HasPrefix(ref, "refs/heads/"):
+					g.commit.Refs.Heads = append(g.commit.Refs.Heads, strings.TrimPrefix(ref, "refs/heads/"))
+				default:
+					g.commit.Refs.Heads = append(g.commit.Refs.Heads, ref)
+				}
+			}
 		}
 	}
 	return g.commit
@@ -243,10 +294,12 @@ func (g *Git) StashCount() int {
 	if g.poshgit || g.stashCount != 0 {
 		return g.stashCount
 	}
+
 	stashContent := g.FileContents(g.rootDir, "logs/refs/stash")
 	if stashContent == "" {
 		return 0
 	}
+
 	lines := strings.Split(stashContent, "\n")
 	g.stashCount = len(lines)
 	return g.stashCount
@@ -264,9 +317,11 @@ func (g *Git) Kraken() string {
 		}
 		g.RawUpstreamURL = g.getRemoteURL()
 	}
+
 	if len(g.Hash) == 0 {
 		g.Hash = g.getGitCommandOutput("rev-parse", "HEAD")
 	}
+
 	return fmt.Sprintf("gitkraken://repolink/%s/commit/%s?url=%s", root, g.Hash, url2.QueryEscape(g.RawUpstreamURL))
 }
 
@@ -279,12 +334,7 @@ func (g *Git) shouldDisplay() bool {
 		return false
 	}
 
-	gitdir, err := g.env.HasParentFilePath(".git", true)
-	if err != nil {
-		if !g.props.GetBool(FetchBareInfo, false) {
-			return false
-		}
-
+	if g.props.GetBool(FetchBareInfo, false) {
 		g.realDir = g.env.Pwd()
 		bare := g.getGitCommandOutput("rev-parse", "--is-bare-repository")
 		if bare == trueStr {
@@ -292,11 +342,10 @@ func (g *Git) shouldDisplay() bool {
 			g.workingDir = g.realDir
 			return true
 		}
-
-		return false
 	}
 
-	if g.shouldIgnoreRootRepository(gitdir.ParentFolder) {
+	gitdir, err := g.env.HasParentFilePath(".git", true)
+	if err != nil {
 		return false
 	}
 
@@ -324,13 +373,21 @@ func (g *Git) setUser() {
 }
 
 func (g *Git) getBareRepoInfo() {
+	// we can still have a pointer to a bare repo
+	if file, err := g.env.HasParentFilePath(".git", true); err == nil && !file.IsDir {
+		content := g.FileContents(file.ParentFolder, ".git")
+		dir := strings.TrimPrefix(content, "gitdir: ")
+		g.workingDir = filepath.Join(file.ParentFolder, dir)
+	}
+
 	head := g.FileContents(g.workingDir, "HEAD")
 	branchIcon := g.props.GetString(BranchIcon, "\uE0A0")
 	g.Ref = strings.Replace(head, "ref: refs/heads/", "", 1)
-	g.HEAD = fmt.Sprintf("%s%s", branchIcon, g.Ref)
+	g.HEAD = fmt.Sprintf("%s%s", branchIcon, g.formatBranch(g.Ref))
 	if !g.props.GetBool(FetchUpstreamIcon, false) {
 		return
 	}
+
 	g.Upstream = g.getGitCommandOutput("remote")
 	if len(g.Upstream) != 0 {
 		g.UpstreamIcon = g.getUpstreamIcon()
@@ -338,11 +395,12 @@ func (g *Git) getBareRepoInfo() {
 }
 
 func (g *Git) setDir(dir string) {
-	dir = runtime.ReplaceHomeDirPrefixWithTilde(g.env, dir) // align with template PWD
+	dir = path.ReplaceHomeDirPrefixWithTilde(dir) // align with template PWD
 	if g.env.GOOS() == runtime.WINDOWS {
 		g.Dir = strings.TrimSuffix(dir, `\.git`)
 		return
 	}
+
 	g.Dir = strings.TrimSuffix(dir, "/.git")
 }
 
@@ -352,8 +410,8 @@ func (g *Git) hasWorktree(gitdir *runtime.FileInfo) bool {
 	content = strings.Trim(content, " \r\n")
 	matches := regex.FindNamedRegexMatch(`^gitdir: (?P<dir>.*)$`, content)
 
-	if matches == nil || len(matches["dir"]) == 0 {
-		g.env.Debug("No matches found, directory isn't a worktree")
+	if len(matches) == 0 {
+		log.Debug("no matches found, directory isn't a worktree")
 		return false
 	}
 
@@ -361,22 +419,25 @@ func (g *Git) hasWorktree(gitdir *runtime.FileInfo) bool {
 	// to the mounted path
 	g.workingDir = g.convertToLinuxPath(matches["dir"])
 
+	// if we don't do this, we will identify the submodule as a worktree
+	isSubmodule := strings.Contains(g.workingDir, "/modules/")
+
 	// in worktrees, the path looks like this: gitdir: path/.git/worktrees/branch
-	// strips the last .git/worktrees part
-	// :ind+5 = index + /.git
-	ind := strings.LastIndex(g.workingDir, ".git/worktrees")
-	if ind > -1 {
+	// rootDir needs to become path/.git
+	// realDir needs to become path
+	ind := strings.LastIndex(g.workingDir, "/worktrees/")
+	if ind > -1 && !isSubmodule {
 		gitDir := filepath.Join(g.workingDir, "gitdir")
-		g.rootDir = g.workingDir[:ind+4]
-		g.realDir = strings.TrimSuffix(g.env.FileContent(gitDir), ".git\n")
+		g.rootDir = g.workingDir[:ind]
+		gitDirContent := g.env.FileContent(gitDir)
+		g.realDir = strings.TrimSuffix(gitDirContent, ".git\n")
 		g.IsWorkTree = true
 		return true
 	}
 
 	// in submodules, the path looks like this: gitdir: ../.git/modules/test-submodule
 	// we need the parent folder to detect where the real .git folder is
-	ind = strings.LastIndex(g.workingDir, ".git/modules")
-	if ind > -1 {
+	if isSubmodule {
 		g.rootDir = resolveGitPath(gitdir.ParentFolder, g.workingDir)
 		// this might be both a worktree and a submodule, where the path would look like
 		// this: path/.git/modules/module/path/worktrees/location. We cannot distinguish
@@ -391,6 +452,7 @@ func (g *Git) hasWorktree(gitdir *runtime.FileInfo) bool {
 			g.IsWorkTree = true
 			return true
 		}
+
 		g.realDir = g.rootDir
 		g.workingDir = g.rootDir
 		return true
@@ -437,6 +499,14 @@ func (g *Git) setBranchStatus() {
 }
 
 func (g *Git) cleanUpstreamURL(url string) string {
+	// Azure DevOps
+	if strings.Contains(url, "dev.azure.com") {
+		match := regex.FindNamedRegexMatch(`^.*@(ssh.)?dev\.azure\.com(:v3)?/(?P<ORGANIZATION>[A-Za-z0-9_-]+)/(?P<PROJECT>[A-Za-z0-9_-]+)/(_git/)?(?P<REPOSITORY>[A-Za-z0-9_-]+)$`, url)
+		if len(match) == 4 {
+			return fmt.Sprintf("https://dev.azure.com/%s/%s/_git/%s", match["ORGANIZATION"], match["PROJECT"], match["REPOSITORY"])
+		}
+	}
+
 	if strings.HasPrefix(url, "http") {
 		return url
 	}
@@ -457,9 +527,9 @@ func (g *Git) cleanUpstreamURL(url string) string {
 	}
 
 	if len(match) != 0 {
-		path := strings.Trim(match["PATH"], "/")
-		path = strings.TrimSuffix(path, ".git")
-		return fmt.Sprintf("https://%s/%s", match["URL"], path)
+		repoPath := strings.Trim(match["PATH"], "/")
+		repoPath = strings.TrimSuffix(repoPath, ".git")
+		return fmt.Sprintf("https://%s/%s", match["URL"], repoPath)
 	}
 
 	// codecommit::region-identifier-id://repo-name
@@ -524,7 +594,7 @@ func (g *Git) setGitStatus() {
 		}
 
 		// map conflicts separately when in a merge or rebase
-		if g.Rebase || g.Merge {
+		if g.Rebase != nil || g.Merge {
 			conflict := "AA"
 			full := status[2:4]
 			if full == conflict {
@@ -538,23 +608,28 @@ func (g *Git) setGitStatus() {
 		g.Working.add(workingCode)
 		g.Staging.add(stagingCode)
 	}
+
 	const (
 		HASH         = "# branch.oid "
 		REF          = "# branch.head "
 		UPSTREAM     = "# branch.upstream "
 		BRANCHSTATUS = "# branch.ab "
 	)
+
 	// firstly assume that upstream is gone
 	g.UpstreamGone = true
 	statusFormats := g.props.GetKeyValueMap(StatusFormats, map[string]string{})
+
 	g.Working = &GitStatus{ScmStatus: ScmStatus{Formats: statusFormats}}
 	g.Staging = &GitStatus{ScmStatus: ScmStatus{Formats: statusFormats}}
+
 	untrackedMode := g.getUntrackedFilesMode()
 	args := []string{"status", untrackedMode, "--branch", "--porcelain=2"}
 	ignoreSubmodulesMode := g.getIgnoreSubmodulesMode()
 	if len(ignoreSubmodulesMode) > 0 {
 		args = append(args, ignoreSubmodulesMode)
 	}
+
 	output := g.getGitCommandOutput(args...)
 	for _, line := range strings.Split(output, "\n") {
 		if strings.HasPrefix(line, HASH) && len(line) >= len(HASH)+7 {
@@ -562,16 +637,19 @@ func (g *Git) setGitStatus() {
 			g.Hash = line[len(HASH):]
 			continue
 		}
+
 		if strings.HasPrefix(line, REF) && len(line) > len(REF) {
 			g.Ref = line[len(REF):]
 			continue
 		}
+
 		if strings.HasPrefix(line, UPSTREAM) && len(line) > len(UPSTREAM) {
 			// status reports upstream, but upstream may be gone (must check BRANCHSTATUS)
 			g.Upstream = line[len(UPSTREAM):]
 			g.UpstreamGone = true
 			continue
 		}
+
 		if strings.HasPrefix(line, BRANCHSTATUS) && len(line) > len(BRANCHSTATUS) {
 			status := line[len(BRANCHSTATUS):]
 			splitted := strings.Split(status, " ")
@@ -584,6 +662,7 @@ func (g *Git) setGitStatus() {
 			g.UpstreamGone = false
 			continue
 		}
+
 		addToStatus(line)
 	}
 }
@@ -603,7 +682,7 @@ func (g *Git) setGitHEADContext() {
 		g.Detached = true
 		g.setPrettyHEADName()
 	} else {
-		head := g.formatHEAD(g.Ref)
+		head := g.formatBranch(g.Ref)
 		g.HEAD = fmt.Sprintf("%s%s", branchIcon, head)
 	}
 
@@ -621,30 +700,48 @@ func (g *Git) setGitHEADContext() {
 			origin = formatDetached()
 		} else {
 			head = strings.Replace(head, "refs/heads/", "", 1)
-			origin = branchIcon + g.formatHEAD(head)
+			origin = branchIcon + g.formatBranch(head)
 		}
 		return origin
 	}
 
+	parseInt := func(file string) int {
+		val, _ := strconv.Atoi(g.FileContents(g.workingDir, file))
+		return val
+	}
+
 	if g.env.HasFolder(g.workingDir + "/rebase-merge") {
-		g.Rebase = true
-		origin := getPrettyNameOrigin("rebase-merge/head-name")
+		head := getPrettyNameOrigin("rebase-merge/head-name")
 		onto := g.getGitRefFileSymbolicName("rebase-merge/onto")
-		onto = g.formatHEAD(onto)
-		step := g.FileContents(g.workingDir, "rebase-merge/msgnum")
-		total := g.FileContents(g.workingDir, "rebase-merge/end")
+		onto = g.formatBranch(onto)
+		current := parseInt("rebase-merge/msgnum")
+		total := parseInt("rebase-merge/end")
 		icon := g.props.GetString(RebaseIcon, "\uE728 ")
-		g.HEAD = fmt.Sprintf("%s%s onto %s%s (%s/%s) at %s", icon, origin, branchIcon, onto, step, total, g.HEAD)
+
+		g.Rebase = &Rebase{
+			HEAD:    head,
+			Onto:    onto,
+			Current: current,
+			Total:   total,
+		}
+
+		g.HEAD = fmt.Sprintf("%s%s onto %s%s (%d/%d) at %s", icon, head, branchIcon, onto, current, total, g.HEAD)
 		return
 	}
 
 	if g.env.HasFolder(g.workingDir + "/rebase-apply") {
-		g.Rebase = true
-		origin := getPrettyNameOrigin("rebase-apply/head-name")
-		step := g.FileContents(g.workingDir, "rebase-apply/next")
-		total := g.FileContents(g.workingDir, "rebase-apply/last")
+		head := getPrettyNameOrigin("rebase-apply/head-name")
+		current := parseInt("rebase-apply/next")
+		total := parseInt("rebase-apply/last")
 		icon := g.props.GetString(RebaseIcon, "\uE728 ")
-		g.HEAD = fmt.Sprintf("%s%s (%s/%s) at %s", icon, origin, step, total, g.HEAD)
+
+		g.Rebase = &Rebase{
+			HEAD:    head,
+			Current: current,
+			Total:   total,
+		}
+
+		g.HEAD = fmt.Sprintf("%s%s (%d/%d) at %s", icon, head, current, total, g.HEAD)
 		return
 	}
 
@@ -668,7 +765,7 @@ func (g *Git) setGitHEADContext() {
 				theirs = g.formatSHA(matches["theirs"])
 			default:
 				headIcon = branchIcon
-				theirs = g.formatHEAD(matches["theirs"])
+				theirs = g.formatBranch(matches["theirs"])
 			}
 			g.HEAD = fmt.Sprintf("%s%s%s into %s", icon, headIcon, theirs, formatDetached())
 			return
@@ -720,15 +817,6 @@ func (g *Git) setGitHEADContext() {
 	g.HEAD = formatDetached()
 }
 
-func (g *Git) formatHEAD(head string) string {
-	maxLength := g.props.GetInt(BranchMaxLength, 0)
-	if maxLength == 0 || len(head) < maxLength {
-		return head
-	}
-	symbol := g.props.GetString(TruncateSymbol, "")
-	return head[0:maxLength] + symbol
-}
-
 func (g *Git) formatSHA(sha string) string {
 	if len(sha) <= 7 {
 		return sha
@@ -753,7 +841,7 @@ func (g *Git) setPrettyHEADName() {
 		if strings.HasPrefix(HEADRef, BRANCHPREFIX) {
 			branchName := strings.TrimPrefix(HEADRef, BRANCHPREFIX)
 			g.Ref = branchName
-			g.HEAD = fmt.Sprintf("%s%s", g.props.GetString(BranchIcon, "\uE0A0"), g.formatHEAD(branchName))
+			g.HEAD = fmt.Sprintf("%s%s", g.props.GetString(BranchIcon, "\uE0A0"), g.formatBranch(branchName))
 			return
 		}
 		// no branch, points to commit
@@ -864,12 +952,12 @@ func (g *Git) getSwitchMode(property properties.Property, gitSwitch, mode string
 
 func (g *Git) repoName() string {
 	if !g.IsWorkTree {
-		return runtime.Base(g.env, g.convertToLinuxPath(g.realDir))
+		return path.Base(g.convertToLinuxPath(g.realDir))
 	}
 
 	ind := strings.LastIndex(g.workingDir, ".git/worktrees")
 	if ind > -1 {
-		return runtime.Base(g.env, g.workingDir[:ind])
+		return path.Base(g.workingDir[:ind])
 	}
 
 	return ""
